# Copyright(c) 2002-2004 Association of Universities for Research in Astronomy, Inc.

include <gemini.h>
include <error.h>
include <imhdr.h>
include <tbset.h>

procedure t_gfcube()

# Reconstruct x/y/lambda datacube from extracted IFU spectra
# Version     Sep  9, 2002  JT  (preliminary release in v1.4)
#             Mar 26, 2004  JT  (update for GMOS-S IFU mapping)
#               - new version requires INSTRUME keyword to be GMOS-N/S
# ----

char  inimage[SZ_FNAME], outimage[SZ_FNAME], outprefix[SZ_FNAME], field[2]
real  ssample

char  logfile[SZ_FNAME]
bool  verbose
int   status

# Local variables
char sbuf[SZ_LINE], sciname[SZ_FNAME], mdfname[SZ_FNAME], outname[SZ_FNAME],
     iphuname[SZ_FNAME], ophuname[SZ_FNAME], imsuffix[SZ_FNAME],
     inst_name[SZ_KWVAL]
pointer sciext, mdfext, outext, inphu, outphu
int fnum[1500], nc, xdim, ydim, ldim, lstart, imexists
real xc[1500], yc[1500], xmin, xmax, ymin, ymax, xsample, ysample

real  clgetr()
pointer immap(), tbtopn()
int imgeti(), imaccess(), strcmp(), gte_blankpar(), gin_ismef(),
    gin_multype(), gin_valid()

begin
	
  # Read task parameters:
  call clgstr("inimage", inimage, SZ_FNAME)
  call clgstr("outimage", outimage, SZ_FNAME)
  call clgstr("outprefix", outprefix, SZ_FNAME)
  ssample = clgetr("ssample")

  # Check that the spatial output sampling is non-zero:
  if (ssample==0.0)
    call error(1, "gfcube: ssample must be > 0")
  
  # Set sampling in x/y individually:
  xsample = ssample
  ysample = ssample
  
  # Make sure a unique input image exists:
  iferr(imexists = imaccess(inimage, READ_ONLY))
    call error(1, "gfcube: inimage is not unique - specify extension")
  if (imexists==NO) call error(1, "gfcube: inimage does not exist")

  # Make sure input name contains file extension:
  call gin_afext(inimage, GEM_DMULTEXT, inimage, YES)
  
  # Complain if input name contains image ext or sec:
  call gin_gsuf(inimage, imsuffix)
  if (strcmp(imsuffix, "")!=0)
	call error(1, "gfcube: inimage contains image extension or section")

  # Make sure input image is a MEF file:
  if (gin_ismef(inimage)==NO)
    call error(1, "gfcube: inimage is not a valid multi-extension FITS file")

  # Determine output name and ensure it is a complete FITS file:
  if (gte_blankpar(outimage)==YES) {
    if (gte_blankpar(outprefix)==YES)
      call error(1, "gfcube: both outimage and outprefix are blank")
	call gin_make(outprefix, inimage, GEM_DMULTEXT, outname)
  }
  else {
	call gin_gsuf(outimage, imsuffix)
    if (strcmp(imsuffix, "")!=0)
      call error(1, "gfcube: outimage contains image extension or section")
	call gin_afext(outimage, GEM_DMULTEXT, outname, NO)
	if(gin_multype(outname)==NO)
      call error(1, "gfcube: outimage type is not valid (try FITS)")
  }

  # Check for problematic output name:
  if (imaccess(outname, READ_ONLY)==YES)
    call error(1, "gfcube: outimage already exists")
  if (gin_valid(outname)==NO)
    call error(1, "gfcube: outimage name is not valid")

  # Generate extension names from file names:
  call gin_psuf(inimage, "[SCI,1]", sciname)
  call gin_psuf(inimage, "[MDF]", mdfname)
  call gin_psuf(inimage, "[0]", iphuname)
  call gin_psuf(outname, "[0,append]", ophuname)
  call gin_psuf(outname, "[SCI,1,append]", outname)

  # Open sci extension:
  sciext = immap(sciname, READ_ONLY, NULL)

  # Check that it's 2D:
  if (imgeti(sciext, "i_naxis") != 2)
    call error(1, "gfcube: input extension [SCI,1] is not 2 dimensional")

  # Open PHU and get instrument name:
  inphu = immap(iphuname, READ_ONLY, NULL)
  iferr(call imgstr(inphu, "INSTRUME", inst_name, SZ_KWVAL))
    inst_name[1] = EOS
  else call gte_unpad(inst_name, inst_name, SZ_KWVAL)
  if (strcmp(inst_name, "GMOS") == 0)
    call strcpy ("GMOS-N", inst_name, SZ_KWVAL)
  if (strcmp(inst_name, "GMOS-N") != 0 && strcmp(inst_name, "GMOS-S") != 0)
    call error(1, "gfcube: header keyword INSTRUME must be GMOS-N or GMOS-S")

  # Open MDF:
  mdfext = tbtopn(mdfname, READ_ONLY, NULL)

  # Read MDF co-ordinates & boundaries:
  call getmap(mdfext, inst_name, fnum, xc, yc, xmin, xmax, ymin, ymax, nc)
	
  # Close MDF ext:
  call tbtclo(mdfext)

  # Check that nrows == number of good fibres in MDF:
  if (IM_LEN(sciext, 2) != nc)
    call error(1, "gfcube: num image rows != num good fibres in MDF")

  # Copy primary header to output:
  outphu = immap(ophuname, NEW_COPY, inphu)
  call imunmap(inphu)
  call imunmap(outphu)

  # Open output datacube:
  outext = immap(outname, NEW_COPY, sciext)

  # Determine where to truncate spectra:
  call getllim(sciext, lstart, ldim)

  # Set output dim/WCS:
  call setdim(sciext, outext, xmin, xmax, ymin, ymax, xsample, ysample,
              lstart, xdim, ydim, ldim)
	
  # Resample input spectra into datacube:
  call recon(sciext, outext, nc, xsample, ysample, xmin, ymin,
             xdim, ydim, ldim, lstart, fnum, xc, yc)
	
  # Close files:
  call imunmap(sciext)
  call imunmap(outext)
	
end


procedure getmap(mdfext, inst_name, fnum, xc, yc, xmin, xmax, ymin, ymax, nc)

# Get co-ordinates & field num for good fibres and return co-ordinate
# boundaries for the big field

# Input instrument name should be GMOS-N or GMOS-S

pointer mdfext
char inst_name[SZ_KWVAL]
int fnum[ARB], nc
real xc[ARB], yc[ARB], xmin, xmax, ymin, ymax

pointer xcol, ycol, becol
int n, nrows, beam, inst
bool initlim

define  GMOS_N        0
define  GMOS_S        1

int tbpsta(), strcmp()

begin

  # Get num rows and col nums:
  nrows = tbpsta(mdfext, TBL_NROWS)
  call tbcfnd(mdfext, "xinst", xcol, 1)
  call tbcfnd(mdfext, "yinst", ycol, 1)
  call tbcfnd(mdfext, "beam", becol, 1)

  # Check instrument name:
  if (strcmp(inst_name, "GMOS-S")==0) inst = GMOS_S
  else inst = GMOS_N
	
  # Loop through table rows & get co-ords for good fibres:
  nc=0
  initlim = true
  for (n=1; n <= nrows; n=n+1) {
    call tbegti(mdfext, becol, n, beam)
    if (beam!=-1) {
      nc = nc+1
      call tbegtr(mdfext, xcol, n, xc[nc])
      call tbegtr(mdfext, ycol, n, yc[nc])
      # Determine field from xc:
      if (inst == GMOS_N && xc[nc] < 30.0 ||
          inst == GMOS_S && xc[nc] > 30.0) {
        fnum[nc] = 1
        # Derive field limits:
        if (initlim) {
          xmin = xc[nc]; xmax = xmin
          ymin = yc[nc]; ymax = ymin
          initlim = false
        } else {
          if (xc[nc] < xmin) xmin = xc[nc]
          else if (xc[nc] > xmax) xmax = xc[nc]
          if (yc[nc] < ymin) ymin = yc[nc]
          else if (yc[nc] > ymax) ymax = yc[nc]
        }
      } # if (field==1)
      else fnum[nc] = 2
    } # end if (beam==1)
  } # end for n

end


procedure setdim(sciext, outext, xmin, xmax, ymin, ymax, xsample, ysample,
                 lstart, xdim, ydim, ldim)

# Calculate and set output image dimensions & WCS

pointer sciext, outext
real xmin, xmax, ymin, ymax, xsample, ysample
int xdim, ydim, ldim, lstart

real cenx, ceny, lcrpix, lcrval, lsample
int n
bool gotkw
char kwbuf[SZ_KWVAL]

int imaccf()
real imgetr()

begin

  # Set dimensionality:
  call imputi(outext, "i_naxis", 3)
  call imastr(outext, "CTYPE1", "LINEAR")
  call imastr(outext, "CTYPE2", "LINEAR")
  call imastr(outext, "CTYPE3", "LAMBDA")

  # Calc / set spatial dimensions from sampling:
  xdim = nint((xmax-xmin) / xsample)
  ydim = nint((ymax-ymin) / ysample)
	
  call imputi(outext, "i_naxis1", xdim)
  call imputi(outext, "i_naxis2", ydim)

  # Calc spatial centre co-ords:
  cenx = 0.5 * (xmin + xmax)
  ceny = 0.5 * (ymin + ymax)
	
  # Round co-ordinate boundaries to match output grid:
  # (and reverse x co-ordinate, to get correct orientation)
#  xmin = cenx - 0.5 * (xdim-1) * xsample
#  xmax = cenx + cenx - xmin
  xmax = cenx - 0.5 * (xdim-1) * xsample  # really xmin
  xmin = cenx + cenx - xmax               # really xmax
  ymin = ceny - 0.5 * (ydim-1) * ysample
  ymax = ceny + ceny - ymin

  # Reverse the sample increment in x:
  xsample = -xsample
  
  # Set spatial increment:
  call imaddr(outext, "CD1_1", xsample)
  call imaddr(outext, "CD2_2", ysample)
  call imaddr(outext, "CDELT1", xsample)
  call imaddr(outext, "CDELT2", ysample)
  
  # Set starting co-ord to centre of first pix:
  call imaddr(outext, "CRPIX1", 1.0)
  call imaddr(outext, "CRVAL1", xmin)
  call imaddr(outext, "CRPIX2", 1.0)
  call imaddr(outext, "CRVAL2", ymin)

  # Set wavelength dim:
  call imaddi(outext, "i_naxis3", ldim)
  call imaddi(outext, "DISPAXIS", 3)
	
  # Get/set wavelength increment:
  ifnoerr(lsample = imgetr(sciext, "CD1_1")) gotkw = true
  else ifnoerr(lsample = imgetr(sciext, "CDELT1")) gotkw = true
  else gotkw = false
  if (gotkw) {
    call imaddr(outext, "CD3_3", lsample)
    call imaddr(outext, "CDELT3", lsample)
  }

  # Get/set wavelength ref pix:
  ifnoerr(lcrval = imgetr(sciext, "CRVAL1"))
    ifnoerr(lcrpix = imgetr(sciext, "CRPIX1")) {
      call imaddr(outext, "CRPIX3", lcrpix-real(lstart)+1.0)
      call imaddr(outext, "CRVAL3", lcrval)
    }

  # Make sure LTM values are 1:
  call imaddr(outext, "LTM1_1", 1.0)
  call imaddr(outext, "LTM2_2", 1.0)
  call imaddr(outext, "LTM3_3", 1.0)
	
  # Delete any extra WCS keywords:
  iferr(call imdelf(outext, "WCSDIM"));
  iferr(call imdelf(outext, "LTM1_2"));
  iferr(call imdelf(outext, "LTM2_1"));
  iferr(call imdelf(outext, "CD1_2"));
  iferr(call imdelf(outext, "CD2_1"));
  iferr(call imdelf(outext, "WAT0_001"));
  iferr(call imdelf(outext, "WAT1_001"));
  iferr(call imdelf(outext, "WAT2_001"));
  iferr(call imdelf(outext, "REFPIX1"));
  iferr(call imdelf(outext, "REFPIX2"));

  # Delete apextract keywords:
  # (need a more efficient way of doing this in final code)
  for (n=1; n < 1000; n=n+1) {
    call sprintf(kwbuf, SZ_KWVAL, "APNUM%d")
     call pargi(n)
    if (imaccf(outext, kwbuf) == YES) call imdelf(outext, kwbuf)
	else break
  }
  for (n=1; true; n=n+1) {
    call sprintf(kwbuf, SZ_KWVAL, "APID%d")
     call pargi(n)
    if (imaccf(outext, kwbuf) == YES) call imdelf(outext, kwbuf)
	else break
  }
  for (n=1000; true; n=n+1) {
    call sprintf(kwbuf, SZ_KWVAL, "AP%d")
     call pargi(n)
    if (imaccf(outext, kwbuf) == YES) call imdelf(outext, kwbuf)
	else break
  }
	
end


procedure recon(sciext, outext, nc, xsample, ysample, x1, y1,
                xdim, ydim, ldim, lstart, fnum, xc, yc)

# Resample the input data onto the datacube

pointer sciext, outext
int nc, xdim, ydim, ldim, lstart, fnum[ARB]
real xsample, ysample, x1, y1, xc[ARB], yc[ARB]

# Local variables
long nspix, ninpix
pointer iwk, wk, xout, yout, vout, xarr, yarr, varr, inbuf, indat, outdat
int x, y, l, n, ncf, istat, stat

pointer imgs2r(), imps3r()

begin

  nspix = xdim * ydim
  ninpix = ldim * nc

  # Get memory for workspace arrays:
  call malloc(iwk, 31*nc+nspix, TY_INT)
  call malloc(wk, 8*nc, TY_REAL)
  call malloc(xout, nspix, TY_REAL)
  call malloc(yout, nspix, TY_REAL)
  call malloc(vout, nspix, TY_REAL)
  call malloc(xarr, nc, TY_REAL)
  call malloc(yarr, nc, TY_REAL)
  call malloc(varr, nc, TY_REAL)
  call malloc(inbuf, ninpix, TY_REAL)

  # Generate output grid co-ords:
  for (y=0; y < ydim; y=y+1)
    for (x=0; x < xdim; x=x+1) {
      Memr[xout+y*xdim+x] = x1 + x*xsample
      Memr[yout+y*xdim+x] = y1 + y*ysample
    }

  # Get input array and copy to safe buffer:
  indat = imgs2r(sciext, lstart, lstart+ldim-1, 1, nc)
  call amovr(Memr[indat], Memr[inbuf], ldim*nc)
	
  # Get output array, all at once for fast i/o:
  outdat = imps3r(outext, 1, xdim, 1, ydim, 1, ldim)

  # Work through datacube, one plane at a time:
  for (l=0; l < ldim; l=l+1) {
    
    # Clear output plane buffer:
    call aclrr(Memr[vout], nspix)
	
    # Extract co-ords & values for field 1 into arrays for interp:
    ncf = 0
    for(n=0; n < nc; n=n+1) {
      if (fnum[n]==1) {

        # Copy pixel & co-ord values to array:
        Memr[varr+ncf] = Memr[inbuf+n*ldim+l]
        Memr[xarr+ncf] = xc[n+1]
        Memr[yarr+ncf] = yc[n+1]
        ncf = ncf+1
      } 
    } # end for (n <= nc)

    # call PDA_IDBVIP(md,ncp,ndp,xd,yd,zd,nip,xi,yi,zi,iwk,wk,istat,status
    istat = 0; stat = 0
    call IDBVIP(1, 4, ncf, Memr[xarr], Memr[yarr], Memr[varr], nspix,
		        Memr[xout], Memr[yout], Memr[vout], Memi[iwk], Memr[wk],
		        istat, stat)
	
    if (istat != 0) {
      call eprintf("WARNING: gfcube: interpolation failed at plane %d\n")
       call pargi(l+1)
    }
		
    # Copy output buffer to image:
    call amovr(Memr[vout], Memr[outdat+l*nspix], nspix)
		
  }
	
  # Free buffers:	
  call mfree(iwk, TY_INT)
  call mfree(wk, TY_REAL)
  call mfree(xout, TY_REAL)
  call mfree(yout, TY_REAL)
  call mfree(vout, TY_REAL)
  call mfree(xarr, TY_REAL)
  call mfree(yarr, TY_REAL)
  call mfree(varr, TY_REAL)
  call mfree(inbuf, TY_REAL)
	
end


procedure getllim(sciext, lstart, ldim)

# Calculate useful wavelength limits for output cube

pointer sciext
int lstart, ldim

# Local variables
pointer indat
int nrows, l, n, ngood, minspec

pointer imgs2r()
int gnu_equalr()

begin

  # Get input dimension in lambda:
  ldim = IM_LEN(sciext, 1)
  nrows = IM_LEN(sciext, 2)

  # Set threshold for inclusion:
  minspec = nint(0.99 * nrows)
	
  # Get input image data:
  indat = imgs2r(sciext, 1, ldim, 1, nrows)

  # Find first wavelength with >99% of spectra:
  for (l=0; l < ldim; l=l+1) {

    # Count non-blank spectra in both fields:
    ngood = 0
    for(n=0; n < nrows; n=n+1)
      if (gnu_equalr(Memr[indat+n*ldim+l], 0.0)==NO)
        ngood = ngood+1
		
    if (ngood > minspec) break
				
  } # end for (l < ldim)

  if (l==ldim) lstart = 1
  else lstart = l+1


  # Find last wavelength element with >99% of spectra:
  for (l=ldim-1; l > -1; l=l-1) {

    # Count non-blank spectra in both fields:
    ngood = 0
    for(n=0; n < nrows; n=n+1)
      if (gnu_equalr(Memr[indat+n*ldim+l], 0.0)==NO)
        ngood = ngood+1
		
    if (ngood > minspec) break
				
  } # end for (l < ldim)

  if (l==-1) ldim = ldim-lstart+1
  else ldim = l-lstart+2

end
